/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <fcntl.h>
#include "is_types.h"

#ifndef lint
static	char	sccsid[] = "@(#)db_lock.c	1.5	12:27:25	1/28/92";
#endif

long	isam_set_ticket();
long	isam_set_ticket_wait();
long	isam_get_ticket();
long	isam_get_ticket_wait();

extern	int	db_errno;

int
__db_lock (file, mode, wait)
file_t	*file;
int	wait;
{
	long	ticket;

	if (mode == O_RDONLY) {
		if (wait)
			ticket = isam_get_ticket_wait (file);
		else
			ticket = isam_get_ticket (file);
	} else {
		if (wait)
			ticket = isam_set_ticket_wait (file);
		else
			ticket = isam_set_ticket (file);
	}
	if (ticket == -1)
		return -1;

	file->f_ticket = ticket;
	return 0;
}

int
db_lock (dbm, filenum, wait)
dbm_t	*dbm;
int	filenum;
int	wait;
{
	int	mode;

	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	if (dbm->d_files[filenum].f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	mode = dbm->d_mode;
	return __db_lock (&dbm->d_files[filenum], mode, wait);
}

int
__db_unlock (file)
file_t	*file;
{
	if (file->f_locked == 0)
		return -1;

	if (isam_end_ticket (file))
		return -1;

	return 0;
}

int
db_unlock (dbm, filenum)
dbm_t	*dbm;
int	filenum;
{
	if (filenum < 0 || filenum >= dbm->d_file_count ||
			dbm->d_files[filenum].f_file_name[0] == '\0')
		return (db_errno = DB_INVALID), -1;

	if (dbm->d_files[filenum].f_schema->s_fields == 0)
		return (db_errno = DB_NOTOPEN), -1;

	return __db_unlock (&dbm->d_files[filenum]);
}
