/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 */

#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#ifdef	CURSESX
#include <cursesX.h>
#else	/* !CURSESX */
#include <curses.h>
#endif /* CURSESX */
#ifdef BSD
#include <sgtty.h>
#define strchr index
#define strrchr rindex
#define memcpy(a, b, c) bcopy((b), (a), (c))
#else
#include <termio.h>
#endif /* BSD */
#include <errno.h>
#include <ctype.h>
#include "is_types.h"
#include "is_proto.h"
#include "patchlevel.h"
#include "isam.h"

#ifndef lint
static	char	sccsid[] = "@(#)dbedit.c	1.14	09:19:37	3/30/92";
#endif

#ifdef BSD
struct	sgttyb	tty_cooked;
#else
struct	termio	tty_cooked;
#endif /* BSD */
int	window_offset = 0;
int	opened_window;
int	screen_ok;
char	*menu_bar =
"Add Begin Change Delete End Find Index Match Next Previous Quit Refresh Scroll";
struct	{
	char	*name;
	int	c1;
	int	c2;
	int	match;
} matches[] = {
	"Less",		'<',	0,	IS_LESS,
	"Less-Equal",	'<',	'=',	IS_LESSEQ,
	"Equal",	'=',	0,	IS_EQUAL,
	"Greater-Equal",'>',	'=',	IS_GREATEREQ,
	"Greater",	'>',	0,	IS_GREATER,
	0,		0,	0,	0
};

#define	MDY_DELIMITER	'/'

long	strtol();
double	strtod();
double	ceil();
double	floor();
dbm_t	*db_open();
int	db_open_file();
long	isam_date2int();
char	*isam_int2date();

char	error_msg[BUFSIZ];

extern	int	db_errno;
extern	char	*db_errors[];

void
error (mesg)
char	*mesg;
{
	if (mesg == 0)
		mesg = db_errors[db_errno];

	move (LINES -1, 0);
	clrtoeol ();
	mvprintw (LINES - 1, 0, "%s", mesg);
	refresh ();
	getch ();
	screen_ok = 0;
}

void
set_index (dbm, filenum)
dbm_t	*dbm;
int	filenum;
{
	file_t	*file = &dbm->d_files[filenum];
	int	i;
	int	y, x;
	field_t	*field;
	char	*name;
	char	buf[MAXFIELDNAME+1];
	int	c;

	if (file->f_index == 0)
		return;

	clear ();
	move (0, 0);

	for (i = 1;i < file->f_schema->s_field_count;i++) {
		field = &file->f_schema->s_fields[i];
		if (field->f_type != T_INDEX)
			continue;

		if (field->f_name[0] == '.')
			continue;

		printw ("%.*s\n", MAXFIELDNAME, field->f_name);
	}
	while (1) {
		i = 0;
		buf[0] = '\0';
		move ((LINES-1), 0);
		clrtoeol ();
		mvaddstr ((LINES-1), 0, "Index: ");
		refresh ();

		while (i < MAXFIELDNAME && (c = getch ()) != EOF
				&& c != '\n' && c != '\r') {
			if (c == '\b') {
				if (i > 0) {
					i--;
					getyx (stdscr, y, x);
					mvaddch (y, x - 1, ' ');
					move (y, x - 1);
					buf[i] = '\0';
					refresh ();
				}
				continue;
			}
			if (! isalnum (c) && c != '_')
				continue;

			buf[i++] = c;
			buf[i] = '\0';
			addch (c);
			refresh ();
		}
		if (i == 0)
			return;

		if (name = db_index_name (dbm, filenum, buf)) {
			if ((i = db_open_index (dbm, filenum, name)) >= 0)
				return;
		}
		sprintf (error_msg, "Unable to locate an index for %s", buf);
		beep ();
		error (error_msg);
	}
}

int
get_match ()
{
	int	i;
	int	c;
	int	match;
	int	c1 = 0,
		c2 = 0;

	while (1) {
		move ((LINES-1), 0);
		clrtoeol ();
		match = -1;

		for (i = 0;matches[i].name;i++) {
			if (matches[i].c1 == c1 && matches[i].c2 == c2) {
				standout ();
				match = matches[i].match;
			}
			printw ("%s", matches[i].name);

			if (matches[i].c1 == c1 && matches[i].c2 == c2)
				standend ();

			addch (' ');
		}
		move ((LINES-1), 0);
		refresh ();

		if ((c1 || c2) && match == -1) {
			c1 = c2;
			c2 = 0;
			continue;
		}
		if ((c = getch ()) < 0) {
			match = -1;
			break;
		}
		if (c == 'g' || c == 'G')
			c = '>';
		else if (c == 'l' || c == 'L')
			c = '<';
		else if (c == 'e' || c == 'E')
			c = '=';

		if (c == '>' || c == '<' || c == '=') {
			if (match == -1) {
				c1 = c;
				c2 = 0;
			} else if (c2 != 0) {
				c1 = c2;
				c2 = c;
			} else if (c1 == 0)
				c1 = c;
			else
				c2 = c;
		} else if (c == '\r' || c == '\n') {
			break;
		} else if (c == '\b') {
			if (c2 == 0)
				c1 = 0;
			else
				c2 = 0;
		} else
			beep ();
	}
	move ((LINES-1), 0);
	clrtoeol ();
	refresh ();

	return match;
}

get_command (commands)
char	*commands;
{
	int	c;

	if ((c = getch ()) < 0)
		return -1;

	if (isupper (c))
		c = tolower (c);

	if (strchr (commands, c))
		return c;

	return -1;
}

draw_menu (menu, command)
char	*menu;
int	command;
{
	char 	*cp;
	int	stand_out = 0;

	if (command != 0 && islower (command))
		command = toupper (command);

	move ((LINES-1), 0);
	clrtoeol ();

	for (cp = menu;*cp;cp++) {
		if (command && command == *cp) {
			standout ();
			stand_out = 1;
		}
		if (stand_out && *cp == ' ') {
			standend ();
			stand_out = 0;
		}
		addch (*cp);
	}
	if (stand_out)
		standend ();
}

int
draw_field (field, line)
field_t	*field;
int	line;
{
	int	j;
	int	x, y;

	mvprintw (line, 0, "%.*s", MAXFIELDNAME, field->f_name);
	addch (' ');
	addch ('[');
	switch (field->f_type) {
		case T_CHAR:
			for (j = 0;j < field->f_size;j++) {
				getyx (stdscr, y, x);
				if (x >= (COLS-2)) {
					addch ('>');
					break;
				}
				addch (' ');
			}
			break;
		case T_SHORT:
			for (j = 0;j < 6;j++) addch (' ');
			break;
		case T_INT:
			if (sizeof (int) == sizeof (short))
				for (j = 0;j < 6;j++) addch (' ');
			else
				for (j = 0;j < 10;j++) addch (' ');
			break;
		case T_LONG:
		case T_SERIAL:
			for (j = 0;j < 10;j++) addch (' ');
			break;
		case T_DATE:
			for (j = 0;j < 10;j++) addch (' ');
			break;
		case T_FLOAT:
			for (j = 0;j < 12;j++) addch (' ');
			break;
		case T_DOUBLE:
			for (j = 0;j < 22;j++) addch (' ');
			break;
		case T_MONEY:
			for (j = 0;j < 17;j++) addch (' ');
			break;
	}
	addch (']');
}

int
draw_fields (file)
file_t	*file;
{
	int	i;
	int	fields;
	field_t	*field;

	for (i = 0;i < (LINES-1);i++) {
		move (i, 0);
		clrtoeol ();
	}
	if (window_offset >= file->f_schema->s_field_count)
		window_offset = 0;

	for (fields = i = 0;fields < (LINES-4) && i + window_offset + 1
			< file->f_schema->s_field_count;i++) {
		field = &file->f_schema->s_fields[i + window_offset + 1];
		if (field->f_type == T_INDEX)
			continue;

		if (field->f_name[0] == '.')
			continue;

		draw_field (field, fields);
		fields++;
	}
}

display_field (file, field, row)
file_t	*file;
field_t	*field;
int	row;
{
	short	s;
	int	i;
	long	l;
	float	f;
	double	d;
	char	*cp;
	int	column;

	if ((column = strlen (field->f_name)) >= MAXFIELDNAME)
		column = MAXFIELDNAME;

	column += 2;

	switch (field->f_type) {
		case T_SHORT:
			memcpy (&s, file->f_data_buffer + field->f_offset,
				sizeof s);
			mvprintw (row, column, "%6d", s);
			break;
		case T_INT:
			memcpy (&i, file->f_data_buffer + field->f_offset,
				sizeof i);
			if (sizeof (int) == sizeof (short))
				mvprintw (row, column, "%6d", i);
			else
				mvprintw (row, column, "%10d", i);
			break;
		case T_LONG:
		case T_SERIAL:
			memcpy (&l, file->f_data_buffer + field->f_offset,
				sizeof l);
			mvprintw (row, column, "%10d", l);
			break;
		case T_DATE:
			memcpy (&l, file->f_data_buffer + field->f_offset,
				sizeof l);
			mvprintw (row, column, "%.10s",
				isam_int2date (l, MDY_DELIMITER));
			break;
		case T_FLOAT:
			memcpy (&f, file->f_data_buffer + field->f_offset,
				sizeof f);
			mvprintw (row, column, "%12.6g", f);
			break;
		case T_DOUBLE:
			memcpy (&d, file->f_data_buffer + field->f_offset,
				sizeof d);
			mvprintw (row, column, "%22.11g", d);
			break;
		case T_MONEY:
			memcpy (&d, file->f_data_buffer + field->f_offset,
				sizeof d);
			mvprintw (row, column, "%17.2f", d / 100.0);
			break;
		case T_CHAR:
			cp = file->f_data_buffer + field->f_offset;
			move (row, column);
			for (i = 0;cp[i] != '\0' && i < field->f_size &&
					column + i < (COLS-2);i++) {
				if (isprint (cp[i]))
					addch (cp[i]);
				else
					addch ('?');
			}
			for (;i < field->f_size && column + i < (COLS-2);i++)
				addch (' ');
			break;
		case T_INDEX:
			return -1;
	}
	return 0;
}

int
display_fields (file)
file_t	*file;
{
	int	i;
	int	fields;
	field_t	*field;

	if (window_offset >= file->f_schema->s_field_count)
		window_offset = 0;

	for (fields = i = 0;fields < 20 && i + window_offset + 1
			< file->f_schema->s_field_count;i++) {
		field = &file->f_schema->s_fields[i + window_offset + 1];
		if (field->f_type == T_INDEX)
			continue;

		if (field->f_name[0] == '.')
			continue;

		if (display_field (file, field, fields))
			return -1;

		fields++;
	}
	refresh ();
	return 0;
}

enter_field (field, data, line)
field_t	*field;
char	*data;
int	line;
{
	double	d;
	float	f;
	short	s;
	long	l;
	int	column;
	char	buf[BUFSIZ];
	char	*ep;
	int	i;
	int	x, y;
	int	c;
	int	len;

	if ((column = strlen (field->f_name)) >= MAXFIELDNAME)
		column = MAXFIELDNAME;

	column += 2;

	switch (field->f_type) {
		case T_CHAR:	len = field->f_size; break;
		case T_SHORT:	len = 6; break;
		case T_INT:	if (sizeof (int) == sizeof (short))
					len = 10;
				else
					len = 6;
				break;
		case T_SERIAL:	return 0;
		case T_DATE:
		case T_LONG:	len = 10; break;
		case T_FLOAT:	len = 12; break;
		case T_MONEY:	len = 17; break;
		case T_DOUBLE:	len = 22; break;
		default:	return -1;
	}
	move (line, column);
	for (i = 0;i < len && i + column < (COLS-2);i++)
		addch (' ');

	move (line, column);
	refresh ();

	i = 0;
	while ((c = getch ()) != EOF && c != '\n' &&
			c != '\r' && i < len && i + column < (COLS-1)) {
		if (c == '\b') {
			if (i == 0)
				continue;

			getyx (stdscr, y, x);
			mvaddch (y, x - 1, ' ');
			move (y, x - 1);
			buf[--i] = '\0';
			refresh ();
			continue;
		}
		addch (c);
		buf[i++] = c;
		buf[i] = 0;
		refresh ();
	}
	if (i == 0)
		return 0;

	switch (field->f_type) {
		case T_SHORT:
do_short:
			s = (short) strtol (buf, &ep, 10);
			if (*ep != '\0')
				return -1;

			memcpy (data, &s, sizeof s);
			mvprintw (line, column, "%6d", s);
			break;
		case T_INT:
			if (sizeof (int) == sizeof (short))
				goto do_short;
			else
				goto do_long;
		case T_LONG:
		case T_SERIAL:
do_long:
			l = strtol (buf, &ep, 10);
			if (*ep != '\0')
				return -1;

			memcpy (data, &l, sizeof l);
			mvprintw (line, column, "%10d", l);
			break;
		case T_DATE:
			if ((l = isam_date2int (buf, MDY_DELIMITER)) == 0)
				return 0;

			memcpy (data, (char *) &l, sizeof l);
			if (! (ep = isam_int2date (l, MDY_DELIMITER)))
				ep = "00/00/00";

			mvprintw (line, column, "%.10s", ep);
			break;
		case T_FLOAT:
			f = strtod (buf, &ep);
			if (*ep != '\0')
				return -1;

			memcpy (data, (char *) &f, sizeof f);
			mvprintw (line, column, "%12.6g", f);
			break;
		case T_DOUBLE:
			d = strtod (buf, &ep);
			if (*ep != '\0')
				return -1;

			memcpy (data, (char *) &d, sizeof d);
			mvprintw (line, column, "%22.11g", d);
			break;
		case T_MONEY:
			d = strtod (buf, &ep);
			if (*ep != '\0')
				return -1;

			if (d > 0)
				d = floor (d * 100.0 + 0.5);
			else
				d = ceil (d * 100.0 - 0.5);

			memcpy (data, (char *) &d, sizeof d);
			mvprintw (line, column, "%17.2f", d / 100.0);
			break;
		case T_CHAR:
#ifdef BSD
			bzero (data, field->f_size);
#else
			memset (data, 0, field->f_size);
#endif /* BSD */
			memcpy (data, buf, i);
			break;
		case T_INDEX:
			return -1;
	}
	return 0;
}

enter_fields (file)
file_t	*file;
{
	int	i;
	int	fields;
	field_t	*field;

	if (window_offset >= file->f_schema->s_field_count)
		window_offset = 0;

	for (fields = i = 0;fields < 20 && i + window_offset + 1
			< file->f_schema->s_field_count;i++) {
		field = &file->f_schema->s_fields[i + window_offset + 1];
		if (field->f_type == T_INDEX)
			continue;

		if (field->f_name[0] == '.')
			continue;

		if (enter_field (field, file->f_data_buffer + field->f_offset, fields))
			return -1;

		if (display_field (file, field, fields))
			return -1;

		fields++;
	}
	refresh ();
	return 0;
}

int
enter_index (file, buf)
file_t	*file;
char	*buf;
{
	char	*cp;
	int	indexnum = file->f_current_index;
	int	i;
	int	max;
	field_t	*field;
	sort_key_t *key;

	if (file->f_index == 0 || indexnum <= 1 ||
			indexnum > file->f_index->i_key_count) {
		error ("No current index");
		return -1;
	}
	key = file->f_index->i_keys + indexnum;

	if (key->k_key_size > BUFSIZ)
		return -1;

	max = key->k_field_count;
	if (! key->k_unique)
		max--;

#ifdef BSD
	bzero (buf, key->k_key_size);
#else
	memset (buf, 0, key->k_key_size);
#endif /* BSD */
	for (i = 0;i < (LINES-1);i++) {
		move (i, 0);
		clrtoeol ();
	}
	for (i = 0;i < max;i++) {
		field = &file->f_schema->s_fields[key->k_fields[i]];
		draw_field (field, i);
	}
	refresh ();
	for (cp = buf, i = 0;i < max;i++) {
		field = &file->f_schema->s_fields[key->k_fields[i]];
		if (enter_field (field, cp, i))
			return -1;

		cp += field->f_size;
	}
	return 0;
}

do_menu (dbm, filenum)
dbm_t	*dbm;
int	filenum;
{
	char	buf[BUFSIZ];
	int	command;
	int	i;
	int	match = IS_EQUAL;
	file_t	*file;
	schema_t *schema;
	field_t	*fields, *field;

	file = &dbm->d_files[filenum];
	schema = file->f_schema;
	fields = schema->s_fields;

	while (1) {
		if (! screen_ok) {
			clear ();

			draw_menu (menu_bar, 0);
			draw_fields (file);
			if (file->f_current_record > 0)
				(void) display_fields (file);

			screen_ok = 1;
		} else {
			draw_menu (menu_bar, 0);
		}
		move ((LINES-1), 0);
		refresh ();

		if ((command = get_command ("abcdefimnpqrs")) < 0) {
			beep ();
			continue;
		} else {
			draw_menu (menu_bar, command);
			move ((LINES-1), 0);
			refresh ();
		}
		switch (command) {
			case 'a':
				if (dbm->d_mode == O_RDONLY) {
					beep ();
					error ("Read-only database");
					break;
				}
#ifdef BSD
				bzero (file->f_data_buffer,
					file->f_record_size);
#else
				memset (file->f_data_buffer, 0,
					file->f_record_size);
#endif /* BSD */
				if (! display_fields (file) &&
						! enter_fields (file)) {
					if (db_write_record (dbm, filenum,
						file->f_data_buffer)) {
						beep ();
						error ("Unable to write record");
					}
				} else if (db_read_record (dbm, filenum, (void *) 0, IS_CURR)) {
					beep ();
					error ("Unable to re-read current record");
				}
				screen_ok = 0;
				break;
			case 'b':
				if (db_rewind (dbm, filenum) ||
					db_read_record (dbm, filenum, (void *) 0, IS_NEXT)) {
					beep ();
					error ("Unable to read from BOF");
				}
				screen_ok = 0;
				break;
			case 'c':
				if (dbm->d_mode == O_RDONLY) {
					beep ();
					error ("Read-only database");
					break;
				}
				if (enter_fields (file) ||
					db_update_record (dbm, filenum,
						file->f_data_buffer)) {
					beep ();
					error ("Unable to update record");
				}
				break;
			case 'd':
				if (dbm->d_mode == O_RDONLY) {
					beep ();
					error ("Read-only database");
					break;
				}
				if (file->f_current_record <= 0 ||
						db_delete_record (dbm, filenum)) {
					beep ();
					error ("Unable to delete record");
				} else
					file->f_current_record = 0;

				screen_ok = 0;
				break;
			case 'e':
				if (db_rewind (dbm, filenum) ||
						db_read_record (dbm, filenum, (void *) 0, IS_PREV)) {
					beep ();
					error ("Unable to read from EOF");
				}
				screen_ok = 0;
				break;
			case 'f':
				if (file->f_current_index <= 1) {
					error ("No index defined.  Use the index command");
					set_index (dbm, filenum);

					if (file->f_current_index <= 1)
						break;
				}
				if (enter_index (file, buf))
					break;

				if (db_read_record (dbm, filenum, (void *) buf, match)) {
					beep ();
					error ("Unable to find matching record");
				}
				screen_ok = 0;
				break;
			case 'i':
				set_index (dbm, filenum);
				screen_ok = 0;
				break;
			case 'm':
				if ((i = get_match ()) != -1)
					match = i;

				break;
			case 'n':
				if (db_read_record (dbm, filenum, (void *) 0, IS_NEXT)) {
					beep ();
					if (db_errno != DB_EOF)
						error ("Unable to read next record");
				} else
					display_fields (file);
				break;
			case 'p':
				if (db_read_record (dbm, filenum, (void *) 0, IS_PREV)) {
					beep ();
					if (db_errno != DB_BOF)
						error ("Unable to read previous record");
				} else
					display_fields (file);
				break;
			case 'q':
				return;
			case 'r':
				if (db_read_record (dbm, filenum, (void *) 0, IS_CURR)) {
					beep ();
					error ("Unable to read current record");
				}
				screen_ok = 0;
				break;
			case 's':
				for (i = 0;i < (LINES/2) && window_offset <
						schema->s_field_count;) {
					field = &fields[window_offset];
					window_offset++;

					if (field->f_type == T_INDEX)
						continue;

					if (field->f_name[0] == '.')
						continue;

					i++;
				}
				if (window_offset >= schema->s_field_count)
					window_offset = 0;

				screen_ok = 0;
		}
	}
}

/*
 * usage - display command line syntax and exit
 */

usage ()
{
	fprintf (stderr, "usage: dbedit { -V |  [ -r ] database [ file ] }\n");
	exit (1);
}

#ifdef	VOIDSIG
void
#else
int
#endif
catch (sig)
int	sig;
{
#ifndef NO_IOCTL
#ifdef BSD
	ioctl (0, TIOCSETP, &tty_cooked);
#else
	ioctl (0, TCSETA, &tty_cooked);
#endif /* BSD */
#endif /* !NO_IOCTL */
	if (opened_window) {
		move ((LINES-1), 0);
		refresh ();
		endwin ();
	}
	exit (sig);
}

#ifdef BSD
beep()
{
   return write(2, "\007", 1);
} 
#endif /* BSD */

/*
 * dbedit - add, change, delete records in a single database file
 */

main (argc, argv)
int	argc;
char	**argv;
{
	dbm_t	*dbm;
	int	filenum;
	int	count;
	int	mode = O_RDWR;
	char	pri_index[MAXFILENAME+2];
	char	buf[BUFSIZ];
	char	*index = (char *) 0;
	extern	int	optind;
	extern	char	*optarg;
	int	flag;
	int	i;

	/*
	 * Get the flags from the command line
	 *
	 *	-V - display version number
	 *	-i - set default index
	 *	-r - set read-only mode
	 */

	while ((flag = getopt (argc, argv, "ri:V")) != EOF) {
		switch (flag) {
			case 'V':
				fprintf (stderr, "dbedit: version %s\n",
					VERSION);
				exit (0);
			case 'i':
				index = optarg;
				break;
			case 'r':
				mode = O_RDONLY;
				break;
			default:
				usage ();
		}
	}

	/* 
	 * Check for a database being named and report the names of
	 * all available files if a file isn't specified.  We open
	 * the database then print out the file names and tell the
	 * user to pick one.
	 */

	if (optind >= argc)
		usage ();

	if (! (dbm = db_open (argv[optind], mode))) {
		sprintf (buf, "dbedit: can't open database %s", argv[1]);
		db_error (buf);
		exit (1);
	}
	if (optind + 1 == argc) {
		printf ("You must select one of the following files:\n\n");
		for (count = i = 0;i < dbm->d_file_count;i++) {
			if (dbm->d_files[i].f_file_name[0] == '\0')
				continue;

#if MAXFILENAME < 20

			/*
			 * Try three columns of 25 characters width each
			 */

			printf ("%-25.*s", MAXFILENAME,
					dbm->d_files[i].f_file_name);

			if (++count % 3 == 0)
				putchar ('\n');
#else

			/*
			 * A single column
			 */

			printf ("%-.*s\n", MAXFILENAME,
					dbm->d_files[i].f_file_name);
#endif
		}
#if MAXFILENAME < 20
		if (count % 3)
			putchar ('\n');
#endif
		exit (0);
	}

	/*
	 * Open the user's file and select the default index
	 */

	if ((filenum = db_open_file (dbm, argv[optind + 1])) < 0) {
		sprintf (buf, "dbedit: can't open file %s", argv[optind + 1]);
		db_error (buf);
		exit (1);
	}
	if (index == (char *) 0) {
		strcat (strcpy (pri_index, "."), argv[optind + 1]);
		index = pri_index;
	}
	if (db_open_index (dbm, filenum, index) == -1) {
		sprintf (buf, "dbedit: can't open index %s", index);
		db_error (buf);
		exit (1);
	}

	/* 
	 * Save the current TTY modes so that they can be restored later
	 * on.  The signal catching routine needs to know that the modes
	 * have been saved so it can restore them if it has to.
	 */

#ifdef BSD
	ioctl (0, TIOCGETP, &tty_cooked);
#else
	ioctl (0, TCGETA, &tty_cooked);
#endif /* BSD */

	/*
	 * Initialize curses and set the operating modes
	 */

	initscr ();
	opened_window = 1;
	nonl ();
#ifdef	BSD
	crmode ();
#else
	cbreak ();
#endif /* BSD */
	noecho ();

	/* 
	 * Set up the signal catchers
	 */

#ifdef	SIGINT
	signal (SIGINT, catch);
#endif
#ifdef	SIGQUIT
	signal (SIGQUIT, catch);
#endif

	/*
	 * Display the menu and start accepting user commands
	 */

	do_menu (dbm, filenum);

	/*
	 * Move to the lower left-hand corner so the user's prompt
	 * won't appear in the middle of the old screen.
	 */

	move ((LINES-1), 0);
	refresh ();
	endwin ();

	/* 
	 * Restore the original modes, give the user a fresh new line
	 * and exit.
	 */

#ifdef BSD
	ioctl (0, TIOCSETP, &tty_cooked);
#else
	ioctl (0, TCSETA, &tty_cooked);
#endif /* BSD */
	putchar ('\n');
	exit (0);
	/*NOTREACHED*/
}
