/*
 * Copyright 1991, 1992, John F. Haugh II
 * All rights reserved.
 *
 * Permission is granted to copy and create derivative works for any
 * non-commercial purpose, provided this copyright notice is preserved
 * in all copies of source code, or included in human readable form
 * and conspicuously displayed on all copies of object code or
 * distribution media.
 *
 *	@(#)is_types.h	1.13	10:32:40	2/26/92
 */

/*
 * Manifest constants for this implementation.
 */

#define	MAXFIELDNAME	32
#define	MAXFILENAME	10
#define	MAXSORTFIELDS	10
#define	MAXINDEXFIELDS	10
#define	MAXBLOCKSIZE	4096
#define	MAXRECSIZE	(4*MAXBLOCKSIZE)
#define	FILEMAGIC	0x66696c65L
#define	DICTMAGIC	0x64696374L

/*
 * This is the structure of each individual field in a database
 * file.  It contains the type of the field, its size for
 * character type fields, and its name.  For index-typed fields
 * the size is the index number that the field applies to.
 */

struct	is_field {
	int	f_type;
	int	f_size;
	int	f_offset;
	char	f_name[MAXFIELDNAME];
};
typedef	struct	is_field field_t;

/*
 * This is the structure of a data dictionary entry in the
 * data dictionary in binary format.
 */

struct	is_dict {
	int	f_fileno;
	int	f_fieldno;
	int	f_fieldtype;
	int	f_fieldsize;
	char	f_fieldfile[MAXFILENAME];
	char	f_fieldname[MAXFIELDNAME];
	short	f_fieldkeys[MAXSORTFIELDS];
};
typedef	struct	is_dict	dict_t;

/*
 * This is the structure of an ISAM file header.  One of these
 * structure appears at the beginning of each ISAM file.
 */

struct	is_head {
	off_t	h_free;
	long	h_ticket;
	long	h_serial;
	long	h_bsize;
	long	h_magic;
};
typedef	struct	is_head	header_t;

/*
 * This is the structure of the schema for the database
 * file.  It contains the fields and their descriptions
 * for the file.
 */

struct	is_schema {
	int	s_field_count;
	field_t	*s_fields;
};
typedef	struct	is_schema schema_t;

/*
 * This is the structure of a file view.  It is possible
 * for a user to request only specific fields be accessed
 * for a given file.
 */

struct	is_view {
	int	v_field_count;		/* number of fields in view */
	int	v_packed;		/* packed or unpacked */
	int	v_size;			/* size of view */
	short	*v_fields;		/* field numbers */
	short	*v_offsets;		/* field offsets in view */
	char	*v_buffer;		/* buffer holding field data */
};
typedef	struct	is_view	view_t;

/*
 * This is a higher-level structure (schema independent format) for file
 * views.  Each structure describes a database field.
 */

struct is_desc {
        char    *desc_name;      	/* Name of the field wanted */
        int     desc_type;		/* Type of field expected */
        int     desc_size;		/* Length of field (IS_CHAR only) */
};
typedef	struct	is_desc	desc_t;

/*
 * This is the structure of sort key.  Each key is an
 * array of fields that is used to sort the key by.
 * The field numbers are OR'd with IS_DESCENDING if the
 * associated field is to be compared in reverse order.
 */

struct	is_sort_key {
	int	k_field_number;
	int	k_field_count;
	int	k_key_size;
	int	k_unique;
	short	k_fields[MAXSORTFIELDS];
};
typedef	struct	is_sort_key	sort_key_t;

/*
 * This is all of the indices for a given file.
 */

struct	is_index {
	int	i_key_count;
	sort_key_t *i_keys;
};
typedef	struct	is_index is_index_t;

/*
 * This is the structure of a database file.
 */

struct	is_file {
	char	f_file_name[MAXFILENAME]; /* name of open file */
	int	f_record_size;		/* size of a record in this file */
	int	f_data_fd;		/* file descriptor for data */
	int	f_key_fd;		/* file descriptor for index */
	char	*f_key_buffer;		/* where key data is read into */
	char	*f_data_buffer;		/* where record data is read into */
	int	f_key_index;		/* index the next 2 fields are for */
	off_t	f_key_block;		/* block in index holding key */
	int	f_key_slot;		/* slot in block holding key */
	int	f_key_match;		/* result of last key comparision */
	long	f_current_record;	/* current record */
	unsigned long f_serial;		/* serial number of current record */
	int	f_current_index;	/* current user index user */
	view_t	*f_current_view;	/* current file view */
	long	f_ticket;		/* last ticket for this file */
	int	f_locked;		/* last set ticket locked this file */
	schema_t *f_schema;		/* list of data fields */
	is_index_t *f_index;		/* list of indices */
	long	f_magic;		/* magic number from file open */
	long	f_bsize;		/* block size from file open */
};
typedef	struct	is_file	file_t;

struct	is_dbm {
	int	d_file_count;
	int	d_mode;
	file_t	d_dictionary;
	file_t	*d_files;
};
typedef	struct	is_dbm	dbm_t;

/*
 * This is the structure of a key file block.  A key block is always
 * the same size as the largest file block, as defined by MAXBLOCKSIZE.
 */

struct	is_key_blk {
	long	kb_prev;
	long	kb_left;
	long	kb_right;
	int	kb_cnt;
	char	kb_data[MAXBLOCKSIZE - (3 * sizeof (long)) - sizeof (int)];
};
typedef	struct	is_key_blk keyblk_t;

/*
 * This is the structure of a display field.  It represents the location
 * of a field on the screen and the data in the record.
 */

struct	is_vfield {
	int	vf_x;
	int	vf_y;
	int	vf_file;
	int	vf_field;
	int	vf_readonly;
};
typedef	struct	is_vfield vfield_t;

/*
 * This describes how two files are joined.
 */

struct	is_join {
	int	j_parent_file;
	int	j_parent_field;
	int	j_child_file;
	int	j_child_field;
};
typedef	struct	is_join join_t;

/*
 * This is the structure of a display.  It consists of two parts - the
 * background image and the field data.
 */

struct	is_display {
	int	d_width;
	int	d_height;
	char	*d_background;
	int	d_field_cnt;
	vfield_t *d_fields;
};
typedef	struct is_display display_t;

/*
 * This is the structure which holds all of the screens for the current
 * session.
 */

struct	is_screen {
	int	s_display_cnt;
	display_t *s_displays;
	int	s_join_cnt;
	join_t	*s_joins;
};
typedef	struct	is_screen screen_t;

/*
 * This is the binary format of each field in the screen
 */

struct is_bfield {
	int	bf_x;
	int	bf_y;
	int	bf_join;
	int	bf_readonly;
	char	bf_name[MAXFIELDNAME];
	char	bf_join_file[MAXFILENAME];
	char	bf_field[MAXFIELDNAME];
	char	bf_file[MAXFILENAME];
};
typedef	struct	is_bfield bfield_t;
#define	bf_join_field bf_name

/*
 * This is the format of each display in binary file format
 */

struct	is_bdisplay {
	int	bd_size;
	int	bd_width;
	int	bd_height;
	char	bd_background[4096];
	int	bd_field_cnt;
	bfield_t bd_fields[1];
};
typedef	struct	is_bdisplay bdisplay_t;
#define	bf_next(f)	(((bfield_t *) f) + 1)

/*
 * This is the format of a binary screen file
 */

struct	is_bscreen {
	int	bs_size;
	char	bs_database[MAXFILENAME];
	int	bs_screen_cnt;
	bdisplay_t bs_screens[1];
};
typedef	struct	is_bscreen bscreen_t;
#define	bd_next(s)	((bdisplay_t *) (((char *) s) + s->bd_size))

/*
 * Constants for key lookup
 */

#define	IS_LESS		1
#define	IS_LESSEQ	2
#define	IS_EQUAL	3
#define	IS_GREATEREQ	4
#define	IS_GREATER	5
#define	IS_NEXT		6
#define	IS_PREV		7
#define	IS_CURR		8

#define	IS_ASCENDING	0
#define	IS_DESCENDING	040000

#define	IS_NONUNIQUE	0
#define	IS_UNIQUE	0100000

#ifdef	__STDC__
#define	P(x)	x
#else
#define	P(x)	()
#endif

/*
 * Define VOIDP to be the type that should be used in (void *).  Some
 * compilers don't grok (pointer to void), so you must use (char *)
 * instead.
 */

#define	VOIDP	void	/**/
/* #define VOIDP	char	/**/

dbm_t	*db_create P((char *));
dbm_t	*db_open P((char *, int));
int	db_create_file P((dbm_t *, char *));
int	db_open_file P((dbm_t *, char *));
int	db_close P((dbm_t *));
int	db_close_file P((dbm_t *, int));
int	db_create_field P((dbm_t *, int, field_t *));
int	db_create_index P((dbm_t *, int, field_t *, sort_key_t *));
int	db_read_record P((dbm_t *, int, VOIDP *, int));
int	db_read_view P((dbm_t *, int, VOIDP *, int));
int	db_write_record P((dbm_t *, int, VOIDP *));
int	db_write_view P((dbm_t *, int, VOIDP *));
int	db_update_record P((dbm_t *, int, VOIDP *));
int	db_update_view P((dbm_t *, int, VOIDP *));
int	db_delete_record P((dbm_t *, int));
int	db_delete_file P((dbm_t *, char *));
int	db_open_field P((dbm_t *, int, char *));
int	db_open_index P((dbm_t *, int, char *));
int	db_rewind P((dbm_t *, int));
VOIDP	*db_view P((dbm_t *, int, int, short *, int));
VOIDP	*db_desc_view P((dbm_t *, int, desc_t *, int));
int	db_lock P((dbm_t *, int, int));
int	db_unlock P((dbm_t *, int));
int	db_rename_field P((dbm_t *, int, char *, char *));
int	db_rename_file P((dbm_t *, int, char *));
int	isam_setbuf P((int));

#undef	P

/*
 * Error codes
 */

#define	DB_NOERR	0		/* <No Error> */
#define	DB_EIO		1		/* I/O Error */
#define	DB_NOTOPEN	2		/* File not open */
#define	DB_INVALID	3		/* Invalid argument */
#define	DB_FILE_EXISTS	4		/* File exists */
#define	DB_FIELD_EXISTS	5		/* Field exists */
#define	DB_NOMEM	6		/* Out of memory */
#define	DB_NOT_EMPTY	7		/* File not empty */
#define	DB_NO_FILE	8		/* File does not exist */
#define	DB_NO_FIELD	9		/* Field does not exist */
#define	DB_CANT_LOCK	10		/* Cannot lock file */
#define	DB_BAD_DB	11		/* Data dictionary is corrupt */
#define	DB_NO_INDEX	12		/* No current index */
#define	DB_NO_RECORD	13		/* No current record */
#define	DB_TOO_MANY	14		/* Too many fields */
#define	DB_BAD_DESC	15		/* Field does not match description */
#define	DB_EXISTS	16		/* Record exists */
#define	DB_NOT_EXIST	17		/* Record does not exist */
#define	DB_NOT_FOUND	18		/* File not found */
#define	DB_BOF		19		/* Beginning of file */
#define	DB_EOF		20		/* End of file */
#define	DB_NO_MATCH	21		/* No matching record */
#define	DB_IS_LOCKED	22		/* File is locked */
